/*
 * Tcl-bindings to access DataTranslation DT2801 Driver
 *
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <tcl.h>
#include <blt.h>

#include "tclDT2801.h"
#include <dt2801.h>
#include <adc.h>
#include <dac.h>

static double dac_fullscale[3] = {10.0, 5.0, 2.5};

static double ADC_Convert(int rawdata, int res, double min, double max)
{
	double v;

	v = rawdata * (max - min)/res  + min;

	return v;
}

static int DAC_Convert(double value, int res, double min, double max)
{
	int v;

	v = floor((value-min)/(max-min) * res + 0.5);
        if (v <0 ) {
		v = 0;
	}
        if (v >= res ) {
		v = res-1;
	}

	return v;
}

static int GetGain(Tcl_Interp* interp, char* str, int* gain)
{
	if ( Tcl_GetInt(interp, str, gain) != TCL_OK) {
		sprintf(interp->result, "wrong arg type for gain");
		return TCL_ERROR;
	}
	if (*gain < 0 || *gain > 3) {
		sprintf(interp->result, "gain (%d) argument out of range(0-3)",
			*gain);
		return TCL_ERROR;
	}
	return TCL_OK;
}

static int SetADCParams(int fd, Dt2801_ADCP_t* adp) {

	int stat = 0;

	if ( (stat = ioctl(fd, DT2801_IOCTL_ADC_SET_PARAMS, &adp)) != 0) {
		return stat;
	}

	return 0;
}


static void DT2801_AD_Imm_Delete(ClientData clientData)
{
	DT2801_AD_Imm *dt2801Ptr = (DT2801_AD_Imm *) clientData;

	close(dt2801Ptr->fd);

	ckfree((char *) dt2801Ptr);
}

static void DT2801_AD_Block_Delete(ClientData clientData)
{
	DT2801_AD_Block *dt2801Ptr = (DT2801_AD_Block *) clientData;

	/* The driver does not stop a running block-commandi
	   when we do a 'close' !!*/
	close(dt2801Ptr->fd);

	ckfree((char *) dt2801Ptr);
}

static void DT2801_DA_Imm_Delete(ClientData clientData)
{
	DT2801_DA_Imm *dt2801Ptr = (DT2801_DA_Imm *) clientData;

	close(dt2801Ptr->fd);

	ckfree((char *) dt2801Ptr);
}

/*
static void DT2801_DA_Block_Delete(ClientData clientData)
{
	DT2801_DA_Block *dt2801Ptr = (DT2801_DA_Block *) clientData;

	close(dt2801Ptr->fd);

	ckfree((char *) dt2801Ptr);
}
*/


static int DT2801_AD_Block_Cmd(ClientData clientData, Tcl_Interp* interp,
			       int argc, char** argv)
{
	int result = TCL_OK;
	int length;
	char c;

	DT2801_AD_Block *dt2801Ptr = (DT2801_AD_Block *) clientData;


	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " option ?arg arg ...?\"",
				 (char *) NULL);
		return TCL_ERROR;
	}

	c = argv[1][0];
	length = strlen(argv[1]);


	if ((c == 's') && (strncmp(argv[1], "setmodifier", length) == 0)) {
		int modval;

		if (argc != 4) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " setmodifier <mod> <val>\"",
					 (char *) NULL);
			goto error;
		}
   
		if ( Tcl_GetBoolean(interp, argv[3], &modval) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for "
				"setmodifier %s", argv[3]);
			return TCL_ERROR;
		}

		length = strlen(argv[2]);
		if ((strncmp(argv[2], "continous", length) == 0)) {
			dt2801Ptr->continous = modval;
		} else if ((strncmp(argv[2], "exttrigger", length) == 0)) {
			dt2801Ptr->exttrigger = modval;
		} else if ((strncmp(argv[2], "extclock", length) == 0)) {
			dt2801Ptr->extclock = modval;
		} else {
			sprintf(interp->result, "unknown modifier\n");
			goto error;
		}

	} else if ((c == 'c') &&
		   (strncmp(argv[1], "clockticks", length) == 0)) {
		int ticks;

		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " clockticks <ticks>\"",
					 (char *) NULL);
			goto error;
		}

		if ( Tcl_GetInt(interp, argv[2], &ticks) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for %s",
				"clockticks");
			return TCL_ERROR;
		}
		dt2801Ptr->clockticks = ticks;

	} else if ((c == 's') &&
		   (strncmp(argv[1], "stop", length) == 0)) {

		int stat;

                if (dt2801Ptr->running != 1) {
				sprintf(interp->result,
					"can't stop, not running\n");
				return TCL_ERROR;
		}

		if (argc != 2) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " stop \"",
					 (char *) NULL);
			goto error;
		}

		if (dt2801Ptr->fd != -1) {
			if ( (stat = ioctl(dt2801Ptr->fd,
					   DT2801_IOCTL_ADC_BLOCKSTOP, -1))
			     != 0) {
				sprintf(interp->result,
					"blockstop failed %d\n", stat);
				return TCL_ERROR;
			}

		}
		dt2801Ptr->running = 0;

	} else if ((c == 's') &&
		   (strncmp(argv[1], "start", length)== 0)){

		int start_channel = 0 , end_channel = 0, gain = 0, conv = 10;
		int stat;

		Dt2801_ADCP_t adp;

                if (dt2801Ptr->running == 1) {
				sprintf(interp->result,
					"can't start, still running\n");
				return TCL_ERROR;
		}

		if (argc != 6) {
			Tcl_AppendResult(interp,
					 "wrong # args: should be \"", argv[0],
					 " start stch endch "
					 "gain conv\"",
					 (char *) NULL);
			goto error;
		}

		if ( Tcl_GetInt(interp, argv[2], &start_channel) != TCL_OK) {
			sprintf(interp->result,
				"wrong arg type for %s", "start_channel");
			return TCL_ERROR;
		}
		if ( Tcl_GetInt(interp, argv[3], &end_channel) != TCL_OK) {
			sprintf(interp->result,
				"wrong arg type for %s", "end_channel");
			return TCL_ERROR;
			}

		if ( GetGain(interp, argv[4], &gain) != TCL_OK) {
			return TCL_ERROR;
		}

		if ( Tcl_GetInt(interp, argv[5], &conv) != TCL_OK) {
			sprintf(interp->result,
				"wrong arg type for %s", "conv");
			return TCL_ERROR;
		}

		adp.gain = gain;
		adp.startch = start_channel;
		adp.endch = end_channel;
		adp.conv = conv;

		dt2801Ptr->conv = conv;


		if ( (stat = ioctl(dt2801Ptr->fd, DT2801_IOCTL_SETCLOCK,
				   dt2801Ptr->clockticks)) != 0) {
			sprintf(interp->result, "setclock failed %d\n", stat);
			close(dt2801Ptr->fd);
			return TCL_ERROR;
		}

		if (dt2801Ptr->exttrigger) {
			if ( (stat = ioctl(dt2801Ptr->fd,
					   DT2801_IOCTL_ADC_SETEXTTRIGGER, 1))
			     != 0) {
				sprintf(interp->result,
					"setexttrig failed %d\n", stat);
				close(dt2801Ptr->fd);
				return TCL_ERROR;
			}
		}

		stat = ioctl(dt2801Ptr->fd, DT2801_IOCTL_ADC_SET_PARAMS,
				   &adp);
		if (stat != 0) {
			sprintf(interp->result, "setadcparams failed %d\n",
				stat);
			return TCL_ERROR;
		}

		stat = ioctl(dt2801Ptr->fd,
			     DT2801_IOCTL_ADC_GET_POLARITY,
			     &dt2801Ptr->polarity);

		if (stat != 0) {
			sprintf(interp->result, "adc_polarity failed %d\n",
				stat);
			return TCL_ERROR;
		}

		if (dt2801Ptr->polarity == 0) {
			dt2801Ptr->min = 0.0;
			dt2801Ptr->max = 10.0/(1<<gain);
		} else if (dt2801Ptr->polarity == 1) {
			double fullscale = 10.0/(1<<gain);
            
			dt2801Ptr->min = -fullscale;
			dt2801Ptr->max = fullscale;
		} else {
			sprintf(interp->result, "bad adc_polarity\n");
			close(dt2801Ptr->fd);
			return TCL_ERROR;
		}

		if ( (stat = ioctl(dt2801Ptr->fd,
				   DT2801_IOCTL_ADC_BLOCKSTART, -1)) != 0) {
			sprintf(interp->result, "blockstart failed %d\n",
				stat);
			return TCL_ERROR;
		}
		dt2801Ptr->running = 1;

		dt2801Ptr->cnt = 0;

		if (dt2801Ptr->ydata.valueArr != NULL) {
			free(dt2801Ptr->ydata.valueArr);
			dt2801Ptr->ydata.valueArr = NULL;
			dt2801Ptr->ydata.numValues = 0;
			dt2801Ptr->ydata.arraySize = 0;
		}

		dt2801Ptr->ydata.numValues = 0;
		dt2801Ptr->ydata.arraySize = conv;
		dt2801Ptr->ydata.valueArr =
			(double*) malloc(sizeof(double) * 
					 dt2801Ptr->ydata.arraySize);

	} else if ((c == 'v') &&
		   (strncmp(argv[1], "vecread", length) == 0)) {

		int stat, ncnt, cnt, i, n;
		static short* data = NULL;
		static int    dsize = 0;

		char* vec = NULL;

		if (argc != 4) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
					 argv[0], " vecread n vector\"",
					 (char *) NULL);
			goto error;
		}

		if ( Tcl_GetInt(interp, argv[2], &n) != TCL_OK) {
			sprintf(interp->result, "wrong arg type for n");
			return TCL_ERROR;
		}

		vec = argv[3]; 

		/* Adjust local buffer if necessary */
		if (n > dsize) {
			if (data != NULL) {
				free(data);
			}
			data = (short *) malloc(n*sizeof(short));
			dsize = n;
		}
         
		ncnt = read(dt2801Ptr->fd, data, n*sizeof(short));


		if (ncnt < 0) {
			sprintf(interp->result, "read failed\n");
			return TCL_ERROR;
		}


		if (ncnt == 0) {
			return TCL_OK;
		}

                cnt = ncnt/sizeof(short);
                /* fprintf(stderr, "cnt = %d\n", cnt); */

		if (dt2801Ptr->raw == 1) {
			for (i=0; i<cnt; i++) {
				dt2801Ptr->ydata.valueArr[i+dt2801Ptr->cnt] =
					data[i];
				/*fprintf(stderr, "%x ", data[i]);*/
			}
			/*			fprintf(stderr, "\n"); */
		} else {
			for (i=0; i<cnt; i++) {
				dt2801Ptr->ydata.valueArr[i+dt2801Ptr->cnt] =
					ADC_Convert(data[i],
						    dt2801Ptr->resolution,
						    dt2801Ptr->min,
						    dt2801Ptr->max);
			}
		}
		dt2801Ptr->ydata.numValues = dt2801Ptr->cnt + cnt;

		if (Blt_ResetVector(interp, vec, &dt2801Ptr->ydata,
				    TCL_STATIC)
		    != TCL_OK) {
			sprintf(interp->result, "ResetVector failed\n");
			return TCL_ERROR;
		}

		dt2801Ptr->cnt += cnt;
		if ( dt2801Ptr->cnt ==  dt2801Ptr->conv) {
			if ( (stat = ioctl(dt2801Ptr->fd,
					   DT2801_IOCTL_ADC_BLOCKSTOP, -1))
			     != 0) {
				sprintf(interp->result,
					"ADC_BLOCKSTOP failed %d\n", stat);
				return TCL_ERROR;
			}
		}
	} else {
		Tcl_AppendResult(interp, "bad option \"", argv[1],
				 "\": must be setmodifier, clockticks,", 
				 " start, vecread or stop",
				 (char *) NULL);
		goto error;
	}
	return result;

error:
	return TCL_ERROR;
}

static int DT2801_AD_Imm_Cmd(ClientData clientData, Tcl_Interp* interp,
			   int argc, char** argv)
{
	DT2801_AD_Imm *dt2801Ptr = (DT2801_AD_Imm *) clientData;
        int gain, stat, cnt, val;


	if (argc != 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " gain\"",
				 (char *) NULL);
		return TCL_ERROR;
	}

	if ( GetGain(interp, argv[1], &gain) != TCL_OK) {
		return TCL_ERROR;
	}

	if ( (stat = ioctl(dt2801Ptr->fd,
			   DT2801_IOCTL_ADC_SET_GAIN, gain)) != 0) {
		sprintf(interp->result, "ADC_SET_GAIN failed %d\n", stat);
		return TCL_ERROR;
	}

	cnt = read(dt2801Ptr->fd, &val, sizeof(val));

	if (cnt != sizeof(val)) {
		sprintf(interp->result, "read failed %d\n", cnt);
		return TCL_ERROR;
	}
/*		fprintf(stderr, "%d\n", val); */

        stat = ioctl(dt2801Ptr->fd, DT2801_IOCTL_ADC_GET_POLARITY,
                             &dt2801Ptr->polarity);

        if (stat != 0) {
                 sprintf(interp->result, "adc_polarity failed %d\n",
                           stat);
                 return TCL_ERROR;
        }

        if (dt2801Ptr->polarity == 0) {
               dt2801Ptr->min = 0.0;
               dt2801Ptr->max = 10.0/(1<<gain);
        } else if (dt2801Ptr->polarity == 1) {
               double fullscale = 10.0/(1<<gain);

               dt2801Ptr->min = -fullscale;
               dt2801Ptr->max = fullscale;
        } else {
               sprintf(interp->result, "bad adc_polarity\n");
               close(dt2801Ptr->fd);
               return TCL_ERROR;
        }


	if (dt2801Ptr->raw == 1) {
		sprintf(interp->result, "%d", val);
	} else {
		sprintf(interp->result, "%f", 
			ADC_Convert(val, dt2801Ptr->resolution,
				    dt2801Ptr->min, dt2801Ptr->max));
	}

	return TCL_OK;
}

static int DT2801_DA_Imm_Cmd(ClientData clientData, Tcl_Interp* interp,
			    int argc, char** argv)
{
	int length;
	char c;
	int stat, cnt;
	double voltage;
	short val;


	DT2801_DA_Imm *dt2801Ptr = (DT2801_DA_Imm *) clientData;


	c = argv[1][0];
	length = strlen(argv[1]);


	if (argc != 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " voltage\"", (char *) NULL);
		return TCL_ERROR;
	}

	if ( Tcl_GetDouble(interp, argv[1], &voltage) != TCL_OK) {
		sprintf(interp->result, "wrong arg type for voltage");
		return TCL_ERROR;
	}

	if (dt2801Ptr->raw == 1) {
		val = voltage;
	} else {
		val = DAC_Convert(voltage, dt2801Ptr->resolution,
				  dt2801Ptr->min, dt2801Ptr->max);
	}

	if ( (stat = ioctl(dt2801Ptr->fd, DT2801_IOCTL_DAC_SET_MODE,
				   DT2801_DAC_IMMEDIATE)) != 0) {
		sprintf(interp->result, "setdacmode failed %d\n", stat);
		return TCL_ERROR;
	}
	
	cnt = write(dt2801Ptr->fd, &val,sizeof(val));

	if (cnt != sizeof(val)) {
		sprintf(interp->result, "write failed %d\n", cnt);
		return TCL_ERROR;
	}
	sprintf(interp->result, "%d\n", val);

	return TCL_OK;
}

static int DT2801_AD_Cmd(ClientData clientData, Tcl_Interp* interp,
			 int argc, char** argv)
{
        int stat, channel;
	char dt2801dev[255]; 

	if (argc != 2 && argc != 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " commandname\" ?channel?",
				 (char *) NULL);
		return TCL_ERROR;
	}

	if (argc == 3) {
		DT2801_AD_Imm* dt2801Ptr;
		
		if ( Tcl_GetInt(interp, argv[2], &channel) != TCL_OK) {
			sprintf(interp->result,
				"wrong arg type for channel");
			return TCL_ERROR;
		}
		sprintf(dt2801dev, "/dev/dt2801/adc%d", channel);
		dt2801Ptr = (DT2801_AD_Imm *)
			ckalloc(sizeof(DT2801_AD_Imm));
		
		/* TCL replaces the 'open'-call  by its own
		   version */
		dt2801Ptr->fd = open(dt2801dev, O_RDONLY, 0000);
		if ( dt2801Ptr->fd < 0 ) {
			/* return perror() instead */
			sprintf(interp->result,
				"Failed to open %s\n", dt2801dev);
			return TCL_ERROR;
		}
		dt2801Ptr->raw = 1;
		dt2801Ptr->bits = 12;
		dt2801Ptr->resolution = 4096;
		dt2801Ptr->polarity = -1; 
		dt2801Ptr->gain = 0; 
		dt2801Ptr->max = 10.0; 
		dt2801Ptr->min = 0.0; 
			
		dt2801Ptr->channel = channel; 

		dt2801Ptr->exttrigger = 0;


		Tcl_CreateCommand(interp, argv[1], DT2801_AD_Imm_Cmd,
				  (ClientData) dt2801Ptr,
				  (Tcl_CmdDeleteProc*)
				  DT2801_AD_Imm_Delete);
		Tcl_SetResult(interp, argv[1], TCL_VOLATILE);
		
	} else if (argc == 2) {
		DT2801_AD_Block* dt2801Ptr;

		sprintf(dt2801dev, "/dev/dt2801/adc");
		dt2801Ptr = (DT2801_AD_Block *)
			ckalloc(sizeof(DT2801_AD_Block));

		/* TCL replaces the 'open'-call  by its own
		   version */
		dt2801Ptr->fd = open(dt2801dev, O_RDONLY, 0000);
		if ( dt2801Ptr->fd < 0 ) {
			/* return perror() instead */
			sprintf(interp->result,
				"Failed to open %s\n", dt2801dev);
			return TCL_ERROR;
		}

		dt2801Ptr->raw = 1;
		dt2801Ptr->bits = 12;
		dt2801Ptr->resolution = 4096;
		dt2801Ptr->polarity = -1; 
		dt2801Ptr->gain = 0; 
		dt2801Ptr->max = 10.0; 
		dt2801Ptr->min = 0.0; 
			
		dt2801Ptr->startch = -1; 
		dt2801Ptr->endch = -1; 

		dt2801Ptr->exttrigger = 0;
		dt2801Ptr->extclock = 0;
		dt2801Ptr->continous = 0;
		dt2801Ptr->clockticks = 250;


		dt2801Ptr->ydata.numValues = 0;
		dt2801Ptr->ydata.arraySize = 0;
		dt2801Ptr->ydata.valueArr = NULL;

		dt2801Ptr->running = 0;

		Tcl_CreateCommand(interp, argv[1], DT2801_AD_Block_Cmd,
				  (ClientData) dt2801Ptr,
				  (Tcl_CmdDeleteProc*)
				  DT2801_AD_Block_Delete);
		Tcl_SetResult(interp, argv[1], TCL_VOLATILE);

	}
	return TCL_OK;
}

static int DT2801_DA_Cmd(ClientData clientData, Tcl_Interp* interp,
			 int argc, char** argv)
{
	DT2801_DA_Imm* dt2801Ptr;
	char dt2801dev[255]; 
	int port, stat;

	if (argc != 3) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
				 argv[0], " commandname\" port",
				 (char *) NULL);
		return TCL_ERROR;
	}
		
	if (Tcl_GetInt(interp, argv[2], &port) != TCL_OK) {
		sprintf(interp->result,
			"wrong arg type for port");
		return TCL_ERROR;
	}
	if (port<0 || port>=2) {  /* XXX */
		sprintf(interp->result, "port out of range (0-2)");
		return TCL_ERROR;
	}

	sprintf(dt2801dev, "/dev/dt2801/dac%d", port);
	dt2801Ptr = (DT2801_DA_Imm *) ckalloc(sizeof(DT2801_DA_Imm));
		
	/* TCL replaces the 'open'-call  by its own
	   version */
	dt2801Ptr->fd = open(dt2801dev, O_WRONLY, 0000);
	if ( dt2801Ptr->fd < 0 ) {
		/* return perror() instead */
		sprintf(interp->result,
			"Failed to open %s\n", dt2801dev);
		return TCL_ERROR;
	}

	dt2801Ptr->raw = 0;

	dt2801Ptr->bits = 12;
	dt2801Ptr->resolution = 4096;

	dt2801Ptr->polarity = -1; 
	dt2801Ptr->fullscale = 10.0; 

	stat = ioctl(dt2801Ptr->fd,
		     DT2801_IOCTL_DAC_GET_POLARITY, &dt2801Ptr->polarity);

	if (stat != 0) {
		sprintf(interp->result, "dac_polarity failed %d\n", stat);
		return TCL_ERROR;
	}

	stat = ioctl(dt2801Ptr->fd,
		     DT2801_IOCTL_DAC_GET_FULLSCALE, &dt2801Ptr->fullscale);

	if (stat != 0) {
		sprintf(interp->result, "dac_fullscale failed %d\n", stat);
		return TCL_ERROR;
	}

	if (dt2801Ptr->polarity == 0) {
		dt2801Ptr->min = 0.0;
		dt2801Ptr->max = dac_fullscale[dt2801Ptr->fullscale];
	} else if (dt2801Ptr->polarity == 1) {
		dt2801Ptr->max = dac_fullscale[dt2801Ptr->fullscale];
		
		dt2801Ptr->min = -dt2801Ptr->max;
	} else {
		sprintf(interp->result, "bad dac_polarity\n");
		close(dt2801Ptr->fd);
		return TCL_ERROR;
	}

	dt2801Ptr->port = port; 
	
	dt2801Ptr->exttrigger = 0;
	
	
	Tcl_CreateCommand(interp, argv[1], DT2801_DA_Imm_Cmd,
			  (ClientData) dt2801Ptr,
			  (Tcl_CmdDeleteProc*)
			  DT2801_DA_Imm_Delete);
	Tcl_SetResult(interp, argv[1], TCL_VOLATILE);
	
	return TCL_OK;
}

static int DT2801_DIO_Cmd(ClientData clientData, Tcl_Interp* interp,
			  int argc, char** argv)
{
	return TCL_OK;
}



int Dt2801_Init(Tcl_Interp *interp)
{
	Tcl_CreateCommand(interp, "dt2801_ad", DT2801_AD_Cmd,
			  (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
	Tcl_CreateCommand(interp, "dt2801_da", DT2801_DA_Cmd,
			  (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
	Tcl_CreateCommand(interp, "dt2801_dio", DT2801_DIO_Cmd,
			  (ClientData) NULL, (Tcl_CmdDeleteProc*) NULL);
	return TCL_OK;
}
